//-------------------------------------------------------------------
// File   : ub_xnet_global.h
// Author : Uwe Becker
// Date   : 2018-01-14
//
// Project : XNET
// Modul   : Global settings
// Global typedefs and network parameter
//-------------------------------------------------------------------

#ifndef UB_XNET_GLOBAL_H_
#define UB_XNET_GLOBAL_H_


//-------------------------------------------------------------------
// attention!!
// use the same settings in this file for all nodes in the XNET network
// independent of the hardware configuration of a single node
//-------------------------------------------------------------------


//-------------------------------------------------------------------
// to port this project to another MCU:
// change the include lines and the typedeclerations
//-------------------------------------------------------------------

// includes for this MCU (stm32f407vgt6)
#include "stm32f4xx.h"
#include <stdbool.h>


// typedeclerations in the xnet project
typedef bool		U8_BOOL;	// boolean type
typedef uint8_t		U8_CHAR;	// 8bit unsigned
typedef uint16_t	U16_WORD;	// 16bit unsigned
typedef uint32_t	U32_DWORD;	// 32bit unsigned

// boolean defines
#define XFALSE		false
#define XTRUE		true



//-------------------------------------------------------------------
// global network parameter
// change this values to your needs
//-------------------------------------------------------------------

//-------------------------------
// network settings
//-------------------------------
#define XNET_MAX_PORTS					4		// maximum ports on a single node (max = 8, the node can have less ports than this number)
#define XNET_MAX_NODES					16		// maximum nodes in the network (max = 240)


//-------------------------------
// frame settings
//-------------------------------
#define XNET_MIN_FRAME_LENGTH			5		// min length of a frame (don't change this value) [LEN,CMD,OWN_ID,CRC16]
#define XNET_MAX_FRAME_LENGTH			25		// max length of a frame (max = 240, make sure UART RX-buffer is big enough)


//-------------------------------
// data settings
//-------------------------------
#define XNET_MIN_DATA_LENGTH			1		// min length of the data area (don't change this value)
#define XNET_MAX_DATA_LENGTH			5		// max length of the data area



//-------------------------------
// connection timing/settings
//-------------------------------
#define XNET_HELLO_DELAY_INTERVAL_ms	300		// interval between two hello packages at one port
#define XNET_OPEN_REQUEST_TIMEOUT_ms	20		// time until neighbor must respond to an open request
#define XNET_OPEN_TX_RETRY				2		// tx open retries until error
#define XNET_PACKAGE_TX_RETRY			3		// tx package retries until error
#define XNET_OPEN_COLLISION_RETRY		3		// retries after collision until error
#define XNET_BYTES_TIMEOUT_ms			5		// timeout for receiving bytes during package transfer
#define XNET_TX_PACKAGE_TTL				(XNET_MAX_NODES)		// time-to-live value for Tx packages

#define XNET_HELLO_PHASE_DURATION		(XNET_HELLO_DELAY_INTERVAL_ms * 3)	// duration of hello-phase after power-up
#define XNET_CONNECTION_PHASE_DURATION	500									// duration of connection phaser after hello-phase
#define XNET_STATUS_LED_ON_TIME_ms		5									// status LED on-time
#define XNET_PING_TIMEOUT_ms			5000	// timeout for ping cmd (5 sec)


//-------------------------------
// UART settings
//-------------------------------
#define XNET_UART_BAUDRATE				115200	// baudrate of uart ports (make sure your UART library can handle this value)
#define XNET_UART_MODE					0		// 0 = 8N1 (no other mode implemented yet)


//-------------------------------
#define XNET_NODE_ID_UNDEFINED			0x00	// undefined node ID (don't change this value)
#define XNET_PORT_NR_UNDEFINED			0xFF	// undefined port_nr (don't change this value)
#define XNET_HOPS_UNDEFINED				0xFF	// undefined hop counts (don't change this value)


//-------------------------------------------------------------------
// don't make changes below this line
//-------------------------------------------------------------------


//-------------------------------
// FRAME identifier list (system commands)
//-------------------------------
#define XNET_FRAME_UNDEFINED			0x00	// undefined frame ID (don't change this value)					[HEADER / PAYLOAD / CRC]
#define XNET_FRAME_HELLO				0x21	// 33d send hello message periodically to direct neighbor		[LEN,CMD,OWN_ID,HOST_ID/CRC16]
#define XNET_FRAME_FLOOD_LINK_STATE		0x22	// 34d flood LinkState message of a single node					[LEN,CMD,OWN_ID,HOST_ID/TTL,P0..Pn/CRC16]
#define XNET_FRAME_REQ_LINK_STATE_TABLE	0x23	// 35d request LinkStateTable from direct neighbor				[LEN,CMD,OWN_ID,HOST_ID/CRC16]
//-------------------------------
// FRAME identifier list (debug commands)
//-------------------------------
#define XNET_FRAME_SEND_DEBUG2_NODE		0x24	// 36d send debug package to a node								[LEN,CMD,OWN_ID,HOST_ID/DEST_ID,MODE,D0..Dn/CRC16]
#define XNET_FRAME_FLOOD_DEBUG2_ALL		0x25	// 37d flood debug package to all nodes							[LEN,CMD,OWN_ID,HOST_ID/TTL,MODE,D0..Dn/CRC16]
#define XNET_FRAME_DEBUG_ANSWER			0x26	// 38d send debug answer back to host of debug					[LEN,CMD,OWN_ID,HOST_ID/DEST_ID,MODE,D0..Dn/CRC16]
//-------------------------------
// FRAME identifier list (user commands)
//-------------------------------
#define XNET_FRAME_SEND_DATA2_NODE		0x27	// 39d send data to a node										[LEN,CMD,OWN_ID,HOST_ID/DEST_ID,MODE,D0..Dn/CRC16]
#define XNET_FRAME_FLOOD_DATA2_GROUP	0x28	// 40d flood data to a group of nodes							[LEN,CMD,OWN_ID,HOST_ID/TTL,GROUP,MODE,D0..Dn/CRC16]
#define XNET_FRAME_DATA_ANSWER			0x29	// 41d send data answer back to host of data					[LEN,CMD,OWN_ID,HOST_ID/DEST_ID,MODE,D0..Dn/CRC16]
//-------------------------------

//-------------------------------
// command list (for package transfer)
//-------------------------------
#define XNET_CMD_UNDEFINED				0x00	// undefined command ID (don't change this value)
#define XNET_CMD_OPEN_REQ				0xF1	// try to open a port for TX package
#define XNET_CMD_OPEN_ACK				0xF2	// open acknowledge
#define XNET_CMD_TX_OPEN				0xF3	// TX port is open
#define XNET_CMD_RX_OPEN				0xF4	// RX port is open
#define XNET_CMD_FRAME_OK				0xF5	// frame is ok
#define XNET_CMD_FRAME_ERROR			0xF6	// error in frame
#define XNET_CMD_TX_CLOSED				0xF7	// TX port is closed
#define XNET_CMD_RX_CLOSED				0xF8	// RX port is closed
#define XNET_CMD_OPEN_COLLISION			0xF9	// collision detected while opening


//---------------------------------
// debug modes (for debug commands)
//---------------------------------
#define XNET_DBG_MODE_PING				0x00	// send/receive ping
#define XNET_DBG_MODE_GET_NODE_VERSION	0x01	// get node version info
#define XNET_DBG_MODE_GET_NODE_ID		0x02	// get NodeId


//---------------------------------
// data modes (for data commands)
//---------------------------------
#define XNET_DATA_MODE_SET_GPIO0		0x00	// set GPIO-0
#define XNET_DATA_MODE_GET_GPIO0		0x01	// get GPIO-0



#if XNET_MAX_PORTS == 0
  #error XNET_MAX_PORTS: nodes must have at least one port
#endif

#if XNET_MAX_PORTS > 8
  #error XNET_MAX_PORTS: too many ports for the nodes
#endif

#if XNET_MAX_NODES < 2
  #error XNET_MAX_NODES: you must have at least 2 nodes in the network
#endif

#if XNET_MAX_NODES > 240
  #error XNET_MAX_NODES: too many nodes for the network
#endif

#if XNET_MIN_FRAME_LENGTH != 5
  #error XNET_MIN_FRAME_LENGTH: min length must be 5
#endif

#if XNET_MAX_FRAME_LENGTH < XNET_MIN_FRAME_LENGTH
  #error XNET_MAX_FRAME_LENGTH: max length must be higher than min lenght
#endif

#if XNET_MAX_FRAME_LENGTH < (XNET_MAX_NODES+5)
  #error XNET_MAX_FRAME_LENGTH: max length must be higher than XNET_MAX_NODES
#endif

#if XNET_MAX_FRAME_LENGTH > 240
  #error XNET_MAX_FRAME_LENGTH: length too high
#endif

#if XNET_PACKAGE_TX_RETRY < 2
  #error XNET_PACKAGE_TX_RETRY: too less retries
#endif

#if XNET_OPEN_TX_RETRY < 2
  #error XNET_OPEN_TX_RETRY: too less retries
#endif




#endif /* UB_XNET_GLOBAL_H_ */
