//-------------------------------------------------------------------
// File    : ub_xnet_l2.h
// Author  : Uwe Becker
// Date    : 2018-01-14
// Version : 0.01
//
// Project : XNET
// Modul   : Protocol layer L2 [network layer]
// Network handling and User API's (for application layer)
//-------------------------------------------------------------------



//-------------------------------------------------------------------
// don't make changes below this line
// to port this project to another MCU or hardware:
// change "xnet_l0.c", "xnet_l0.h", "xnet_global.h"
//-------------------------------------------------------------------



// include own header file
#include "ub_xnet_l2.h"




// internal functions
void check_l2_flags(void);
void new_own_link_state(void);
void update_own_link_state(void);
void flood_own_link_state(void);
void flood_neighbor_link_state(void);
void calculate_path_table(void);
U8_BOOL is_node_id_in_NodeIdList(U8_CHAR node_id);
U8_CHAR get_port_for_node(U8_CHAR node_id);
void transfer_package(U8_CHAR frame_id);
void flood_data2_group(void);
void flood_debug2_all(void);
U8_CHAR get_rx_buf_cnt(void);
void copy_own_package(void);
void inc_buf_rd_ptr(void);
void inc_buf_wr_ptr(void);
U8_BOOL flood_link_state_entry(void);
U8_CHAR get_connected_node_index(U8_CHAR item_nr);
void request_link_state_table(void);
void check_own_debug_package(void);
void check_own_debug_answer_package(void);

void add_link_state_entry(void);
void clear_all_tables(void);
U8_CHAR convert_node_id(U8_CHAR node_id);
U8_BOOL is_node_in_my_network(U8_CHAR node_id);




// init xnet layer-2
// must be called once after power-up from user to init this node
// own_id = unique network ID (<> 0x00)
// return:
// XTRUE = init ok
U8_BOOL UB_Xnet_L2_Init(U8_CHAR own_id)
{
	U8_CHAR n, m;

	if(own_id == XNET_NODE_ID_UNDEFINED) return XFALSE;

	// clear other data
	XNet.connection.nr_nodes = 0;
	XNet.connection.nr_neighbors = 0;
	XNet.group_mask = 0x01;

	XNet.state.path_table_ok = XFALSE;
	XNet.state.new_neighbor = XFALSE;
	XNet.state.neighbor_port = XNET_PORT_NR_UNDEFINED;
	XNet.state.flooding_table = XFALSE;
	XNet.state.flood_ls_node_idx = 0;

	XNet.rx_buf.rd_ptr = 0;
	XNet.rx_buf.wr_ptr = 0;
	for(m = 0; m < NODE_RX_BUFFER_SIZE; m++) {
		XNet.rx_buf.payload[m].frame_id = 0;
		XNet.rx_buf.payload[m].port_nr = 0;
		XNet.rx_buf.payload[m].host_id = XNET_NODE_ID_UNDEFINED;
		XNet.rx_buf.payload[m].data_length = 0;
		for(n = 0; n < XNET_MAX_FRAME_LENGTH; n++) {
			XNet.rx_buf.payload[m].data[n] = 0x00;
		}
	}

	XNet.ping.running = XFALSE;
	XNet.ping.pong_ok = XFALSE;
	XNet.ping.start_time = 0;
	XNet.ping.stop_time = 0;


	// init xnet layer-1
	UB_Xnet_L1_Init(own_id);

	// clear all tables of the node
	clear_all_tables();

	return XTRUE;
}


// 1ms Tick
// must be called every ms from user (use a timer or the systick to do this)
void UB_Xnet_L2_1ms_Tick(void)
{
	UB_Xnet_L1_1ms_Tick();
}



// must be call periodically from user (e.g. from main loop)
void UB_Xnet_L2_Loop(void)
{
	static U32_DWORD loop_cnt_old = 0;
	U8_CHAR frame_id;


	// process this function only once every ms
	if(Node.cnt.loop_cnt == loop_cnt_old) return;
	loop_cnt_old = Node.cnt.loop_cnt;

	// call layer-1 loop to transmit and receive frames
	UB_Xnet_L1_Loop();


	if(Node.l2_flags.flags != L2_FLAG_NONE) {
		check_l2_flags();
	}

	if(UB_Xnet_L1_Check_Rx_Buffer() == XTRUE) {
		// get frame_id from rx_payload
		frame_id = Node.rx_payload.frame_id;

		switch(frame_id) {
			case XNET_FRAME_FLOOD_LINK_STATE:
				// link state from another node received
				if((Node.rx_payload.host_id != XNET_NODE_ID_UNDEFINED) && (Node.rx_payload.host_id != Node.my_id)) {
					// only check packages from other nodes
					add_link_state_entry();
					// send link state to all other ports
					flood_neighbor_link_state();
					// set flag to calculate new path table
					Node.l2_flags.flags |= L2_FLAG_CALC_NEW_PATH_TABLE;
					Node.l2_flags.CALC_NEW_PATH_TABLE_cnt = 0;
				}
			break;
			case XNET_FRAME_REQ_LINK_STATE_TABLE:
				// LinkStateTable request from neighbor received
				if((Node.rx_payload.host_id != XNET_NODE_ID_UNDEFINED) && (Node.rx_payload.host_id != Node.my_id)) {
					// only check packages from other nodes

					// set flag to flood complete LinkStateTable to new neighbor
					Node.l2_flags.flags |= L2_FLAG_FLOOD_COMPLETE_LS_TABLE;
					XNet.state.flooding_table = XTRUE;
					XNet.state.flood_ls_node_idx = 1;
					XNet.state.neighbor_port = Node.rx_payload.port_nr;
				}
			break;
			case XNET_FRAME_SEND_DEBUG2_NODE:
				// debug package for a node received
				if((Node.rx_payload.host_id != XNET_NODE_ID_UNDEFINED) && (Node.rx_payload.host_id != Node.my_id)) {
					// only check packages from other nodes
					if(Node.rx_payload.data[0] == Node.my_id) {
						// package is for me so check what to do
						check_own_debug_package();
					}
					else {
						// package is not for me, transfer to destination
						transfer_package(XNET_FRAME_SEND_DEBUG2_NODE);
					}
				}
			break;
			case XNET_FRAME_FLOOD_DEBUG2_ALL:
				// debug package for a node received
				if((Node.rx_payload.host_id != XNET_NODE_ID_UNDEFINED) && (Node.rx_payload.host_id != Node.my_id)) {
					// only check packages from other nodes
					check_own_debug_package();
					// send debug package to all other ports
					flood_debug2_all();
				}
			break;
			case XNET_FRAME_DEBUG_ANSWER:
				// debug answer package for a node received
				if((Node.rx_payload.host_id != XNET_NODE_ID_UNDEFINED) && (Node.rx_payload.host_id != Node.my_id)) {
					// only check packages from other nodes
					if(Node.rx_payload.data[0] == Node.my_id) {
						// package is for me so check what to do
						check_own_debug_answer_package();
					}
					else {
						// package is not for me, transfer to destination
						transfer_package(XNET_FRAME_DEBUG_ANSWER);
					}
				}
			break;
			case XNET_FRAME_SEND_DATA2_NODE:
				// data package for a node received
				if((Node.rx_payload.host_id != XNET_NODE_ID_UNDEFINED) && (Node.rx_payload.host_id != Node.my_id)) {
					// only check packages from other nodes
					if(Node.rx_payload.data[0] == Node.my_id) {
						// package is for me so copy in rx buf
						copy_own_package();
					}
					else {
						// package is not for me, transfer to destination
						transfer_package(XNET_FRAME_SEND_DATA2_NODE);
					}
				}
			break;
			case XNET_FRAME_FLOOD_DATA2_GROUP:
				// data package for a group received
				if((Node.rx_payload.host_id != XNET_NODE_ID_UNDEFINED) && (Node.rx_payload.host_id != Node.my_id)) {
					// only check packages from other nodes
					if((Node.rx_payload.data[1] & XNet.group_mask) > 0) {
						// package is for my group so copy in rx buf
						copy_own_package();
					}
					// send data package to all other ports
					flood_data2_group();
				}
			break;
			case XNET_FRAME_DATA_ANSWER:
				// data answer for a node received
				if((Node.rx_payload.host_id != XNET_NODE_ID_UNDEFINED) && (Node.rx_payload.host_id != Node.my_id)) {
					// only check packages from other nodes
					if(Node.rx_payload.data[0] == Node.my_id) {
						// package is for me so copy in rx buf
						copy_own_package();
					}
					else {
						// package is not for me, transfer to destination
						transfer_package(XNET_FRAME_DATA_ANSWER);
					}
				}
			break;
			default:
				// nothing to do here
				loop_cnt_old = 0;
		}
	}
}

// check if rx buffer is empty or not
// return:
// XTRUE = rx buffer is not empty
// XFALSE = rx buffer is empty
U8_BOOL UB_Xnet_L2_Check_Rx_Buffer(void)
{
	if(get_rx_buf_cnt() > 0) {
		return XTRUE;
	}

	return XFALSE;
}

// read data from rx buffer
// return: type of received command
// CMD_TYPE_EMPTY
// CMD_TYPE_UNKNOWN
// CMD_TYPE_DEBUG_ANSWER
// CMD_TYPE_NODE_DATA
// CMD_TYPE_GROUP_DATA
// return values:
// source_id = NodeId of the message host
// mode = mode type
// data_ptr = pointer to the received data
// data_size = length of the received data
CmdType_t UB_Xnet_L2_Read_Rx_Buffer(U8_CHAR *source_id, U8_CHAR *mode, U8_CHAR *data_ptr, U8_CHAR *data_size)
{
	U8_CHAR frame_id, pos;
	U8_CHAR lenght, n;

	*source_id = XNET_NODE_ID_UNDEFINED;
	*data_size = 0;

	if(get_rx_buf_cnt() > 0) {
		pos = XNet.rx_buf.rd_ptr;
		frame_id = XNet.rx_buf.payload[pos].frame_id;
		inc_buf_rd_ptr();

		switch(frame_id) {
			case XNET_FRAME_DEBUG_ANSWER:
				*source_id = XNet.rx_buf.payload[pos].host_id;
				*mode = XNet.rx_buf.payload[pos].data[1];
				lenght = XNet.rx_buf.payload[pos].data_length-2;
				*data_size = lenght;
				for(n = 0; n < lenght; n++) {
					data_ptr[n] = XNet.rx_buf.payload[pos].data[n+1];
				}
				return CMD_TYPE_DEBUG_ANSWER;
			break;
			case XNET_FRAME_SEND_DATA2_NODE:
				*source_id = XNet.rx_buf.payload[pos].host_id;
				*mode = XNet.rx_buf.payload[pos].data[1];
				lenght = XNet.rx_buf.payload[pos].data_length-2;
				*data_size = lenght;
				for(n = 0; n < lenght; n++) {
					data_ptr[n] = XNet.rx_buf.payload[pos].data[n+2];
				}
				return CMD_TYPE_NODE_DATA;
			break;
			case XNET_FRAME_FLOOD_DATA2_GROUP:
				*source_id = XNet.rx_buf.payload[pos].host_id;
				*mode = XNet.rx_buf.payload[pos].data[2];
				lenght = XNet.rx_buf.payload[pos].data_length-3;
				*data_size = lenght;
				for(n = 0; n < lenght; n++) {
					data_ptr[n] = XNet.rx_buf.payload[pos].data[n+3];
				}
				return CMD_TYPE_GROUP_DATA;
			break;
			case XNET_FRAME_DATA_ANSWER:
				*source_id = XNet.rx_buf.payload[pos].host_id;
				*mode = XNet.rx_buf.payload[pos].data[1];
				lenght = XNet.rx_buf.payload[pos].data_length-2;
				*data_size = lenght;
				for(n = 0; n < lenght; n++) {
					data_ptr[n] = XNet.rx_buf.payload[pos].data[n+2];
				}
				return CMD_TYPE_DATA_ANSWER;
			break;
			default:
				return CMD_TYPE_UNKNOWN;
		}
	}

	return CMD_TYPE_EMPTY;
}

// readout a XNet Info
U32_DWORD UB_Xnet_L2_Get_NetInfo(NetInfo_t mode)
{
	U32_DWORD dword;

	switch(mode) {
		case NETINFO_NR_NODES: return XNet.connection.nr_nodes; break;
		case NETINFO_NR_NEIGHBORS: return XNet.connection.nr_neighbors;break;
		case NETINFO_OWN_NODE_ID: return Node.my_id; break;
		case NETINFO_OWN_GROUP_MASK: return XNet.group_mask; break;
		case NOTINFO_VERSION_NR: // Layer2,Layer1,Layer0,Hardware
			dword = (XNET_VERSION_L2_8BIT << 24) | (XNET_VERSION_L1_8BIT << 16) | (XNET_VERSION_L0_8BIT << 8) | (XNET_VERSION_HW_8BIT);
			return dword;
		break;
	}

	return 0;
}

// returns the NodeId [n] in the current XNet network
// node_index : [0..NETINFO_NR_NODES]
U8_CHAR UB_Xnet_L2_Get_Node_ID(U8_CHAR node_index)
{
	U8_CHAR idx;

	if(node_index > XNet.connection.nr_nodes) return XNET_NODE_ID_UNDEFINED;

	idx = get_connected_node_index(node_index);
	return XNet.Node_Id_List[idx];
}

// returns the number of hops to reach the Node [n]
// node_index : [0..NETINFO_NR_NODES]
U8_CHAR UB_Xnet_L2_Get_Neighbor_Hops(U8_CHAR node_index)
{
	U8_CHAR idx;

	if(node_index > XNet.connection.nr_nodes) return XNET_HOPS_UNDEFINED;

	idx = get_connected_node_index(node_index);
	return XNet.Path_Table[idx].hop_cnt;
}

// returns the current group mask of the node
void UB_Xnet_L2_Set_Group_Mask(U8_CHAR group_mask)
{
	XNet.group_mask = group_mask;
}

// send debug package to another node in the XNet network
// destination_id = NodeId of the destination Node
// dbg_mode = debug mode
// data_ptr = pointer to the data (NULL if data_size = 0)
// data_size = length of the data (0..n)
U8_BOOL UB_Xnet_L2_Send_Debug_To_Node(U8_CHAR destination_id, U8_CHAR dbg_mode, U8_CHAR *data_ptr, U8_CHAR data_size)
{
	U8_CHAR port_nr, port_mask, n;

	if(Node.state != NODE_STATE_IDLE) return XFALSE;
	if(data_size > XNET_MAX_DATA_LENGTH) return XFALSE;
	if(XNet.connection.nr_nodes == 1) return XFALSE;
	if(is_node_in_my_network(destination_id) == XFALSE) return XFALSE;

	port_nr = get_port_for_node(destination_id);
	if(Node.port[port_nr].connected == XFALSE) return XFALSE;
	if(UB_Xnet_L1_Tx_Buffer_Full() == XTRUE) return XFALSE;
	// copy package header
	Node.tx_payload.host_id = Node.my_id; // source
	Node.tx_payload.data_length = data_size+2;
	Node.tx_payload.data[0] = destination_id; // destination
	Node.tx_payload.data[1] = dbg_mode; // mode
	// copy package data
	if(data_size > 0) {
		for(n = 0; n < data_size; n++) {
			Node.tx_payload.data[n+2] = data_ptr[n];
		}
	}
	port_mask = UB_Xnet_L1_Convert_Port_Nr_In_Port_Mask(port_nr, PORTMASK_SINGLE_PORT);
	UB_Xnet_L1_Add_Tx_Package(port_mask, XNET_FRAME_SEND_DEBUG2_NODE);

	return XTRUE;
}

// send debug package to all other nodes in the XNet network
// dbg_mode = debug mode
// data_ptr = pointer to the data (NULL if data_size = 0)
// data_size = length of the data (0..n)
U8_BOOL UB_Xnet_L2_Send_Debug_To_All(U8_CHAR dbg_mode, U8_CHAR *data_ptr, U8_CHAR data_size)
{
	U8_CHAR port_mask, n;

	if(Node.state != NODE_STATE_IDLE) return XFALSE;
	if(data_size > XNET_MAX_DATA_LENGTH) return XFALSE;
	if(XNet.connection.nr_nodes == 1) return XFALSE;
	if(XNet.connection.nr_neighbors == 0) return XFALSE;
	if(UB_Xnet_L1_Tx_Buffer_Full() == XTRUE) return XFALSE;

	// copy package header
	Node.tx_payload.host_id = Node.my_id; // source
	Node.tx_payload.data_length = data_size+2;
	Node.tx_payload.data[0] = XNET_TX_PACKAGE_TTL; // TTL
	Node.tx_payload.data[1] = dbg_mode; // mode
	// copy package data
	if(data_size > 0) {
		for(n = 0; n < data_size; n++) {
			Node.tx_payload.data[n+2] = data_ptr[n];
		}
	}
	port_mask = UB_Xnet_L1_Convert_Port_Nr_In_Port_Mask(0, PORTMASK_ANY_PORT);
	UB_Xnet_L1_Add_Tx_Package(port_mask, XNET_FRAME_FLOOD_DEBUG2_ALL);

	return XTRUE;
}

// send data package to another node in the XNet network
// destination_id = NodeId of the destination Node
// data_mode = data mode
// data_ptr = pointer to the data (NULL if data_size = 0)
// data_size = length of the data (0..n)
U8_BOOL UB_Xnet_L2_Send_Data_To_Node(U8_CHAR destination_id, U8_CHAR data_mode, U8_CHAR *data_ptr, U8_CHAR data_size)
{
	U8_CHAR port_nr, port_mask, n;

	if(Node.state != NODE_STATE_IDLE) return XFALSE;
	if(data_size > XNET_MAX_DATA_LENGTH) return XFALSE;
	if(XNet.connection.nr_nodes == 1) return XFALSE;
	if(is_node_in_my_network(destination_id) == XFALSE) return XFALSE;

	port_nr = get_port_for_node(destination_id);
	if(Node.port[port_nr].connected == XFALSE) return XFALSE;
	if(UB_Xnet_L1_Tx_Buffer_Full() == XTRUE) return XFALSE;
	// copy package header
	Node.tx_payload.host_id = Node.my_id; // source
	Node.tx_payload.data_length = data_size+2;
	Node.tx_payload.data[0] = destination_id; // destination
	Node.tx_payload.data[1] = data_mode; // mode
	// copy package data
	if(data_size > 0) {
		for(n = 0; n < data_size; n++) {
			Node.tx_payload.data[n+2] = data_ptr[n];
		}
	}
	port_mask = UB_Xnet_L1_Convert_Port_Nr_In_Port_Mask(port_nr, PORTMASK_SINGLE_PORT);
	UB_Xnet_L1_Add_Tx_Package(port_mask, XNET_FRAME_SEND_DATA2_NODE);

	return XTRUE;
}

// send data package to all nodes in the XNet network (Nodes with matching group mask are receiving the data)
// group_flags = flags for corresponding GroupMask of the nodes
// data_mode = data mode
// data_ptr = pointer to the data (NULL if data_size = 0)
// data_size = length of the data (0..n)
U8_BOOL UB_Xnet_L2_Send_Data_To_Group(U8_CHAR group_flags, U8_CHAR data_mode, U8_CHAR *data_ptr, U8_CHAR data_size)
{
	U8_CHAR port_mask, n;

	if(Node.state != NODE_STATE_IDLE) return XFALSE;
	if(data_size > XNET_MAX_DATA_LENGTH) return XFALSE;
	if(XNet.connection.nr_nodes == 1) return XFALSE;
	if(XNet.connection.nr_neighbors == 0) return XFALSE;
	if(UB_Xnet_L1_Tx_Buffer_Full() == XTRUE) return XFALSE;

	// copy package header
	Node.tx_payload.host_id = Node.my_id; // source
	Node.tx_payload.data_length = data_size+3;
	Node.tx_payload.data[0] = XNET_TX_PACKAGE_TTL; // TTL
	Node.tx_payload.data[1] = group_flags; // group flags
	Node.tx_payload.data[2] = data_mode; // mode
	// copy package data
	if(data_size > 0) {
		for(n = 0; n < data_size; n++) {
			Node.tx_payload.data[n+3] = data_ptr[n];
		}
	}
	port_mask = UB_Xnet_L1_Convert_Port_Nr_In_Port_Mask(0, PORTMASK_ANY_PORT);
	UB_Xnet_L1_Add_Tx_Package(port_mask, XNET_FRAME_FLOOD_DATA2_GROUP);

	return XTRUE;
}

// send answer package to another node in the XNet network
// destination_id = NodeId of the destination Node
// data_mode = data mode
// data_ptr = pointer to the data (NULL if data_size = 0)
// data_size = length of the data (0..n)
U8_BOOL UB_Xnet_L2_Send_Answer_To_Node(U8_CHAR destination_id, U8_CHAR data_mode, U8_CHAR *data_ptr, U8_CHAR data_size)
{
	U8_CHAR port_nr, port_mask, n;

	if(Node.state != NODE_STATE_IDLE) return XFALSE;
	if(data_size > XNET_MAX_DATA_LENGTH) return XFALSE;
	if(XNet.connection.nr_nodes == 1) return XFALSE;
	if(is_node_in_my_network(destination_id) == XFALSE) return XFALSE;

	port_nr = get_port_for_node(destination_id);
	if(Node.port[port_nr].connected == XFALSE) return XFALSE;
	if(UB_Xnet_L1_Tx_Buffer_Full() == XTRUE) return XFALSE;
	// copy package header
	Node.tx_payload.host_id = Node.my_id; // source
	Node.tx_payload.data_length = data_size+2;
	Node.tx_payload.data[0] = destination_id; // destination
	Node.tx_payload.data[1] = data_mode; // mode
	// copy package data
	if(data_size > 0) {
		for(n = 0; n < data_size; n++) {
			Node.tx_payload.data[n+2] = data_ptr[n];
		}
	}
	port_mask = UB_Xnet_L1_Convert_Port_Nr_In_Port_Mask(port_nr, PORTMASK_SINGLE_PORT);
	UB_Xnet_L1_Add_Tx_Package(port_mask, XNET_FRAME_DATA_ANSWER);

	return XTRUE;
}

// get current TIC value (ms)
U32_DWORD UB_Xnet_L2_Get_Tic(void)
{
	return (Node.cnt.time_cnt);
}

// send ping message to another node in the XNet network
// destination_id = NodeId of the destination Node
U8_BOOL UB_Xnet_L2_Start_Ping_Node(U8_CHAR destination_id)
{
	U8_CHAR port_nr, port_mask;

	if(XNet.ping.running != XFALSE) return XFALSE;
	if(Node.state != NODE_STATE_IDLE) return XFALSE;
	if(XNet.connection.nr_nodes == 1) return XFALSE;
	if(is_node_in_my_network(destination_id) == XFALSE) return XFALSE;

	port_nr = get_port_for_node(destination_id);
	if(Node.port[port_nr].connected == XFALSE) return XFALSE;
	if(UB_Xnet_L1_Tx_Buffer_Full() == XTRUE) return XFALSE;
	// copy packet header
	Node.tx_payload.host_id = Node.my_id; // source
	Node.tx_payload.data_length = 2;
	// copy packet data
	Node.tx_payload.data[0] = destination_id; // destination
	Node.tx_payload.data[1] = XNET_DBG_MODE_PING; // mode
	port_mask = UB_Xnet_L1_Convert_Port_Nr_In_Port_Mask(port_nr, PORTMASK_SINGLE_PORT);
	UB_Xnet_L1_Add_Tx_Package(port_mask, XNET_FRAME_SEND_DEBUG2_NODE);

	XNet.ping.running = XTRUE;
	XNet.ping.pong_ok = XFALSE;
	XNet.ping.start_time = Node.cnt.time_cnt;
	XNet.ping.stop_time = 0;

	return XTRUE;
}


// increment read pointer
void inc_buf_rd_ptr(void)
{
	XNet.rx_buf.rd_ptr++;
	if(XNet.rx_buf.rd_ptr >= NODE_RX_BUFFER_SIZE) XNet.rx_buf.rd_ptr = 0;
}

// increment write pointer
void inc_buf_wr_ptr(void)
{
	XNet.rx_buf.wr_ptr++;
	if(XNet.rx_buf.wr_ptr >= NODE_RX_BUFFER_SIZE) XNet.rx_buf.wr_ptr = 0;
	if(XNet.rx_buf.wr_ptr == XNet.rx_buf.rd_ptr) Node.dbg_cnt.buf_ovf_cnt++;
}

// copy received package in rx buffer
void copy_own_package(void)
{
	U8_CHAR n,pos;

	pos = XNet.rx_buf.wr_ptr;

	XNet.rx_buf.payload[pos].frame_id = Node.rx_payload.frame_id;
	XNet.rx_buf.payload[pos].data_length = Node.rx_payload.data_length;
	XNet.rx_buf.payload[pos].port_nr = Node.rx_payload.port_nr;
	XNet.rx_buf.payload[pos].host_id = Node.rx_payload.host_id;
	// toDo: change array length to a smaller value data < frame
	for(n = 0; n < XNET_MAX_FRAME_LENGTH; n++) {
		XNet.rx_buf.payload[pos].data[n] = Node.rx_payload.data[n];
	}
	inc_buf_wr_ptr();
}

// debug package received
void check_own_debug_package(void)
{
	U8_CHAR dbg_mode, port_nr, port_mask, destination_id;
	U8_CHAR data_size;

	destination_id = Node.rx_payload.host_id; // source is new destination
	if(is_node_in_my_network(destination_id) == XFALSE) return;

	port_nr = get_port_for_node(destination_id);
	if(Node.port[port_nr].connected == XFALSE) return;

	data_size = Node.rx_payload.data_length;

	dbg_mode = Node.rx_payload.data[1];
	switch(dbg_mode) {
		case XNET_DBG_MODE_PING:
			// create ping package
			Node.tx_payload.host_id = Node.my_id; // own ID
			Node.tx_payload.data_length = 2;
			// copy package data
			Node.tx_payload.data[0] = destination_id; // destination
			Node.tx_payload.data[1] = XNET_DBG_MODE_PING; // mode
			// send package back to host node
			port_mask = UB_Xnet_L1_Convert_Port_Nr_In_Port_Mask(port_nr, PORTMASK_SINGLE_PORT);
			UB_Xnet_L1_Add_Tx_Package(port_mask, XNET_FRAME_DEBUG_ANSWER);
		break;
		case XNET_DBG_MODE_GET_NODE_VERSION:
			// create data package with node infos
			Node.tx_payload.host_id = Node.my_id; // own ID
			Node.tx_payload.data_length = 6;
			// copy package data
			Node.tx_payload.data[0] = destination_id; // destination
			Node.tx_payload.data[1] = XNET_DBG_MODE_GET_NODE_VERSION; // mode
			Node.tx_payload.data[2] = XNET_VERSION_L2_8BIT;
			Node.tx_payload.data[3] = XNET_VERSION_L1_8BIT;
			Node.tx_payload.data[4] = XNET_VERSION_L0_8BIT;
			Node.tx_payload.data[5] = XNET_VERSION_HW_8BIT;
			// send package back to host node
			port_mask = UB_Xnet_L1_Convert_Port_Nr_In_Port_Mask(port_nr, PORTMASK_SINGLE_PORT);
			UB_Xnet_L1_Add_Tx_Package(port_mask, XNET_FRAME_DEBUG_ANSWER);
		break;
		case XNET_DBG_MODE_GET_NODE_ID:
			Node.tx_payload.host_id = Node.my_id; // own ID
			Node.tx_payload.data_length = 2;
			// copy package data
			Node.tx_payload.data[0] = destination_id; // destination
			Node.tx_payload.data[1] = XNET_DBG_MODE_GET_NODE_ID; // mode
			// send package back to host node
			port_mask = UB_Xnet_L1_Convert_Port_Nr_In_Port_Mask(port_nr, PORTMASK_SINGLE_PORT);
			UB_Xnet_L1_Add_Tx_Package(port_mask, XNET_FRAME_DEBUG_ANSWER);
		break;
	}
}

void check_own_debug_answer_package(void)
{
	U8_CHAR dbg_mode;

	dbg_mode = Node.rx_payload.data[1];
	switch(dbg_mode) {
		case XNET_DBG_MODE_PING:
			XNet.ping.running = XFALSE;
			XNet.ping.pong_ok = XTRUE;
			XNet.ping.stop_time = Node.cnt.time_cnt;
			copy_own_package();
		break;
		default:
			copy_own_package();
	}
}


// returns number of packages in rx buf (0 = empty)
U8_CHAR get_rx_buf_cnt(void)
{
	U8_CHAR cnt = 0;

	if(XNet.rx_buf.wr_ptr >= XNet.rx_buf.rd_ptr) {
		cnt = XNet.rx_buf.wr_ptr - XNet.rx_buf.rd_ptr;
	}
	else {
		cnt = (XNet.rx_buf.wr_ptr + NODE_RX_BUFFER_SIZE) - XNet.rx_buf.rd_ptr;
	}

	return cnt;
}

// returns port to reach NodeId
// returns XNET_PORT_NR_UNDEFINED if NodeId is not reachable
U8_CHAR get_port_for_node(U8_CHAR node_id)
{
	U8_CHAR port_nr = XNET_PORT_NR_UNDEFINED, node_idx;

	if(is_node_id_in_NodeIdList(node_id) == XTRUE) {
		// get node_index
		node_idx = convert_node_id(node_id);
		port_nr = XNet.Path_Table[node_idx].port_nr;
	}

	return port_nr;
}


// check the layer-2 flags
void check_l2_flags(void)
{
	// tx buffer must not be full
	if(UB_Xnet_L1_Tx_Buffer_Full() == XTRUE) return;

	if((Node.l2_flags.flags & L2_FLAG_OWN_LINK_STATE_CHANGED) != 0) {
		// own LinkState changed
		Node.l2_flags.OWN_LINK_STATE_CHANGED_cnt++;
		if(Node.l2_flags.OWN_LINK_STATE_CHANGED_cnt > 5) {
			Node.l2_flags.OWN_LINK_STATE_CHANGED_cnt = 0;
			Node.l2_flags.flags &= ~L2_FLAG_OWN_LINK_STATE_CHANGED;
			// update own LinkState
			new_own_link_state();
		}
	}
	else if((Node.l2_flags.flags & L2_FLAG_FLOOD_OWN_LINK_STATE) != 0) {
		// flood own LinkState
		if((Node.state == NODE_STATE_CONNECTION) || (Node.state == NODE_STATE_IDLE)) {
			Node.l2_flags.flags &= ~L2_FLAG_FLOOD_OWN_LINK_STATE;
			flood_own_link_state();
			// set flag to calculate new path table
			Node.l2_flags.flags |= L2_FLAG_CALC_NEW_PATH_TABLE;
			Node.l2_flags.CALC_NEW_PATH_TABLE_cnt = 0;
		}
	}
	else if((Node.l2_flags.flags & L2_FLAG_CALC_NEW_PATH_TABLE) != 0) {
		// calculate PathTable
		if(Node.state == NODE_STATE_IDLE) {
			Node.l2_flags.CALC_NEW_PATH_TABLE_cnt++;
			if(Node.l2_flags.CALC_NEW_PATH_TABLE_cnt > 50) {
				Node.l2_flags.CALC_NEW_PATH_TABLE_cnt = 0;
				Node.l2_flags.flags &= ~L2_FLAG_CALC_NEW_PATH_TABLE;
				// calculate new path table
				calculate_path_table();
				if(XNet.state.path_table_ok == XTRUE) {
					if(XNet.state.new_neighbor == XTRUE) {
						XNet.state.new_neighbor = XFALSE;
						// set flag to flood complete LinkStateTable to new neighbor
						Node.l2_flags.flags |= L2_FLAG_FLOOD_COMPLETE_LS_TABLE;
						XNet.state.flooding_table = XTRUE;
						XNet.state.flood_ls_node_idx = 1;
					}
				}
				else {
					XNet.state.path_table_ok = XTRUE;
					XNet.state.new_neighbor = XFALSE;
				}
			}
		}
	}
	else if((Node.l2_flags.flags & L2_FLAG_FLOOD_COMPLETE_LS_TABLE) != 0) {
		// flood entry from LinkStateTable
		if(flood_link_state_entry() == XTRUE) {
			Node.l2_flags.flags &= ~L2_FLAG_FLOOD_COMPLETE_LS_TABLE;
			if(Node.my_id > Node.port[XNet.state.neighbor_port].neighbor_id) {
				// set flag to request neighbor LinkStateTable
				Node.l2_flags.flags |= L2_FLAG_REQ_COMPLETE_LS_TABLE;
			}
			else {
				XNet.state.flooding_table = XFALSE;
			}
		}
	}
	else if((Node.l2_flags.flags & L2_FLAG_REQ_COMPLETE_LS_TABLE) != 0) {
		// request LinkStateTable from direct neighbor
		Node.l2_flags.flags &= ~L2_FLAG_REQ_COMPLETE_LS_TABLE;
		request_link_state_table();
		XNet.state.flooding_table = XFALSE;
	}
}


// clear complete PathTable
void clear_path_table(void)
{
	U8_CHAR n;

	// clear PathTable
	for(n = 0; n < XNET_MAX_NODES; n++) {
		XNet.Path_Table[n].port_nr = XNET_PORT_NR_UNDEFINED;
		XNet.Path_Table[n].hop_cnt = XNET_HOPS_UNDEFINED;
	}
}

// clear complete NodeIdList
void clear_node_id_list(void)
{
	U8_CHAR n;

	// add this node to the NodeIdList
	XNet.connection.nr_nodes = 1;
	XNet.Node_Id_List[0] = Node.my_id;

	// clear rest of NodeIdList
	for(n = 1; n < XNET_MAX_NODES; n++) {
		XNet.Node_Id_List[n] = XNET_NODE_ID_UNDEFINED;
	}
}

// clear LinkStateTable
void clear_link_state_table(void)
{
	U8_CHAR n, m;

	// clear LinkStateTable
	for(n = 0; n < XNET_MAX_NODES; n++) {
		for(m = 0; m < XNET_MAX_PORTS; m++) {
			XNet.Link_State_Table[n][m] = XNET_NODE_ID_UNDEFINED;
		}
	}
}

// clear NodeIdList, LinkStateTable and PathTable
void clear_all_tables(void)
{
	clear_node_id_list();

	clear_link_state_table();

	clear_path_table();
}


// check if NodeId is already in NodeIdList
U8_BOOL is_node_id_in_NodeIdList(U8_CHAR node_id)
{
	U8_CHAR n;

	// search NodeId in NodeIdList
	for(n = 0; n < XNET_MAX_NODES; n++) {
		if(XNet.Node_Id_List[n] == node_id) return XTRUE;
	}

	return XFALSE;
}

// check if NodeId is reachable
U8_BOOL is_node_in_my_network(U8_CHAR node_id)
{

	if(get_port_for_node(node_id) == XNET_PORT_NR_UNDEFINED) return XFALSE;

	return XTRUE;
}

// add node_id to NodeIdList
U8_BOOL add_node_id(U8_CHAR node_id)
{
	U8_CHAR n;

	for(n = 1; n < XNET_MAX_NODES; n++) {
		if(XNet.Node_Id_List[n] == XNET_NODE_ID_UNDEFINED) {
			// add new node in first free place of NodeIdList
			XNet.Node_Id_List[n] = node_id;
			return XTRUE;
		}
	}

	return XFALSE;
}

// convert node_id into node_idx
U8_CHAR convert_node_id(U8_CHAR node_id)
{
	U8_CHAR n;

	// search NodeId in NodeIdList
	for(n = 0; n < XNET_MAX_NODES; n++) {
		if(XNet.Node_Id_List[n] == node_id) return n;
	}

	return 0;
}

// set link states from neighbor to own LinkStateTable
void set_link_states(U8_CHAR node_id)
{
	U8_CHAR node_idx;
	U8_CHAR n;

	// get node_index
	node_idx = convert_node_id(node_id);
	if(node_idx > 0) {
		// update all port entries of the neighbor node from rx_payload
		for(n = 0; n < XNET_MAX_PORTS; n++) {
			XNet.Link_State_Table[node_idx][n] = Node.rx_payload.data[n+1];
		}
	}
}

// add LinkState entry from neighbor to own NodeIdList and LinkStateTable
void add_link_state_entry(void)
{
	U8_CHAR node_id;
	U8_BOOL status = XTRUE;

	// get node_id from rx_payload
	node_id = Node.rx_payload.host_id;

	// check if this a new node_id
	if(is_node_id_in_NodeIdList(node_id) == XFALSE) {
		// add new NodeId into NodeIdList
		status = add_node_id(node_id);
	}

	if(status == XTRUE) {
		// if node_id ok, update LinkStateTable
		set_link_states(node_id);
	}
}


// transfer the received package to destination
void transfer_package(U8_CHAR frame_id)
{
	U8_CHAR destination_id, port_nr, data_size;
	U8_CHAR port_mask, n;

	destination_id = Node.rx_payload.data[0]; // destination
	if(is_node_in_my_network(destination_id) == XFALSE) return;

	port_nr = get_port_for_node(destination_id);
	if(Node.port[port_nr].connected == XFALSE) return;

	// copy package header
	Node.tx_payload.host_id = Node.rx_payload.host_id;  // copy node_id from received package
	Node.tx_payload.data_length = Node.rx_payload.data_length;
	// copy package data
	data_size = Node.tx_payload.data_length;
	for(n = 0; n < data_size; n++) {
		Node.tx_payload.data[n] = Node.rx_payload.data[n];
	}
	port_mask = UB_Xnet_L1_Convert_Port_Nr_In_Port_Mask(port_nr, PORTMASK_SINGLE_PORT);
	UB_Xnet_L1_Add_Tx_Package(port_mask, frame_id);
}

// flood the received data2 group package to all other neighbors
void flood_data2_group(void)
{
	U8_CHAR port_mask, n;
	U8_CHAR ttl, data_size;

	/// check TTL
	ttl = Node.rx_payload.data[0];
	if(ttl > 0) {
		// add data2 group package from neighbor for all ports (except the rx neighbor) to tx framebuf
		// copy package header
		Node.tx_payload.host_id = Node.rx_payload.host_id;  // copy node_id from received package
		Node.tx_payload.data_length = Node.rx_payload.data_length;
		Node.tx_payload.data[0] = (ttl-1); // decrement TTL
		// copy package data
		data_size = Node.tx_payload.data_length;
		for(n = 1; n < data_size; n++) {
			Node.tx_payload.data[n] = Node.rx_payload.data[n];
		}
		port_mask = UB_Xnet_L1_Convert_Port_Nr_In_Port_Mask(Node.rx_payload.port_nr, PORTMASK_ANY_OTHER_PORT);
		UB_Xnet_L1_Add_Tx_Package(port_mask, XNET_FRAME_FLOOD_DATA2_GROUP);
	}
}

// flood the received debug package to all other neighbors
void flood_debug2_all(void)
{
	U8_CHAR port_mask, n;
	U8_CHAR ttl, data_size;

	/// check TTL
	ttl = Node.rx_payload.data[0];
	if(ttl > 0) {
		// add debug2all package from neighbor for all ports (except the rx neighbor) to tx framebuf
		// copy package header
		Node.tx_payload.host_id = Node.rx_payload.host_id;  // copy node_id from received package
		Node.tx_payload.data_length = Node.rx_payload.data_length;
		Node.tx_payload.data[0] = (ttl-1); // decrement TTL
		// copy package data
		data_size = Node.tx_payload.data_length;
		for(n = 1; n < data_size; n++) {
			Node.tx_payload.data[n] = Node.rx_payload.data[n];
		}
		port_mask = UB_Xnet_L1_Convert_Port_Nr_In_Port_Mask(Node.rx_payload.port_nr, PORTMASK_ANY_OTHER_PORT);
		UB_Xnet_L1_Add_Tx_Package(port_mask, XNET_FRAME_FLOOD_DEBUG2_ALL);
	}
}

// flood the received link state package to all other neighbors
void flood_neighbor_link_state(void)
{
	U8_CHAR port_mask, n;
	U8_CHAR ttl;

	/// check TTL
	ttl = Node.rx_payload.data[0];
	if(ttl > 0) {
		// add LinkState entry from neighbor for all ports (except the rx neighbor) to tx framebuf
		// copy package header
		Node.tx_payload.host_id = Node.rx_payload.host_id;  // copy node_id from received package
		Node.tx_payload.data_length = XNET_MAX_PORTS+1;
		Node.tx_payload.data[0] = (ttl-1); // decrement TTL
		// copy LinkState
		for(n = 0; n < XNET_MAX_PORTS; n++) {
			Node.tx_payload.data[n+1] = Node.rx_payload.data[n+1];
		}
		port_mask = UB_Xnet_L1_Convert_Port_Nr_In_Port_Mask(Node.rx_payload.port_nr, PORTMASK_ANY_OTHER_PORT);
		UB_Xnet_L1_Add_Tx_Package(port_mask, XNET_FRAME_FLOOD_LINK_STATE);
	}
}

// flood own link state package to all neighbors
void flood_own_link_state(void)
{
	U8_CHAR port_mask, n;

	// add LinkState entry 0 from LinkStateTable for all ports to tx framebuf
	// copy package header
	Node.tx_payload.host_id = Node.my_id; // use own ID
	Node.tx_payload.data_length = XNET_MAX_PORTS+1;
	Node.tx_payload.data[0] = XNET_TX_PACKAGE_TTL; // TTL
	// copy LinkState
	for(n = 0; n < XNET_MAX_PORTS; n++) {
		Node.tx_payload.data[n+1] = XNet.Link_State_Table[0][n];
	}
	port_mask = UB_Xnet_L1_Convert_Port_Nr_In_Port_Mask(0, PORTMASK_ANY_PORT);
	UB_Xnet_L1_Add_Tx_Package(port_mask, XNET_FRAME_FLOOD_LINK_STATE);
}

// search connected node (item_nr) and return index nr
U8_CHAR get_connected_node_index(U8_CHAR item_nr)
{
	U8_CHAR n, connected_node = 0, node_index = 0;

	if(item_nr == 0) return 0; // own index

	for(n = 1; n < XNET_MAX_NODES; n++) {
		if(XNet.Path_Table[n].hop_cnt != XNET_HOPS_UNDEFINED) {
			// this is a connected node
			connected_node++;
			if(connected_node == item_nr) {
				node_index = n;
				break;
			}
		}
	}

	return node_index;
}

// request the LinkStateTable from direct neighbor
void request_link_state_table(void)
{
	U8_CHAR port_nr, port_mask;

	port_nr = XNet.state.neighbor_port;
	if((port_nr != XNET_PORT_NR_UNDEFINED) && (Node.port[port_nr].connected == XTRUE)) {
		// request LinkStateTable from neighbor
		Node.tx_payload.host_id = Node.my_id; // source
		Node.tx_payload.data_length = 0;
		port_mask = UB_Xnet_L1_Convert_Port_Nr_In_Port_Mask(port_nr, PORTMASK_SINGLE_PORT);
		UB_Xnet_L1_Add_Tx_Package(port_mask, XNET_FRAME_REQ_LINK_STATE_TABLE);
	}
}

// flood single entry of LinkStateTable to new connected neighbor
U8_BOOL flood_link_state_entry(void)
{
	U8_CHAR port_nr, port_mask, idx, n;

	if(XNet.state.flood_ls_node_idx >= XNet.connection.nr_nodes) {
		// complete LinkStateTable flooded
		return XTRUE;
	}
	else {
		port_nr = XNet.state.neighbor_port;

		if((port_nr != XNET_PORT_NR_UNDEFINED) && (Node.port[port_nr].connected == XTRUE)) {
			idx = get_connected_node_index(XNet.state.flood_ls_node_idx);
			if(XNet.Path_Table[idx].port_nr != port_nr) {
				// don't send neighbor his own connected Nodes
				Node.tx_payload.host_id = XNet.Node_Id_List[idx]; // copy node_id
				Node.tx_payload.data_length = XNET_MAX_PORTS+1;
				Node.tx_payload.data[0] = XNET_TX_PACKAGE_TTL; // TTL
				// copy LinkState
				for(n = 0; n < XNET_MAX_PORTS; n++) {
					Node.tx_payload.data[n+1] = XNet.Link_State_Table[idx][n];
				}
				port_mask = UB_Xnet_L1_Convert_Port_Nr_In_Port_Mask(port_nr, PORTMASK_SINGLE_PORT);
				UB_Xnet_L1_Add_Tx_Package(port_mask, XNET_FRAME_FLOOD_LINK_STATE);
			}
		}
	}

	// next entry
	XNet.state.flood_ls_node_idx++;
	if(XNet.state.flood_ls_node_idx >= XNet.connection.nr_nodes) {
		// complete LinkStateTable flooded
		return XTRUE;
	}

	return XFALSE;
}

// add an entry to the PathTable
U8_BOOL add_path_table_entry(U8_CHAR node_id, U8_CHAR port_nr, U8_CHAR hop_cnt)
{
	U8_CHAR node_idx;

	// get node_index
	node_idx = convert_node_id(node_id);
	if((node_idx > 0) && (XNet.Path_Table[node_idx].hop_cnt == XNET_HOPS_UNDEFINED)) {
		// only set fastest route
		XNet.Path_Table[node_idx].port_nr = port_nr;
		XNet.Path_Table[node_idx].hop_cnt = hop_cnt;

		return XTRUE;
	}

	return XFALSE;
}

// calculate own PathTable
void calculate_path_table(void)
{
	U8_CHAR n, m, node_id, hops, port_nr;
	U8_BOOL next_hop = XFALSE;
	U8_CHAR hop_cnt = 0;

	// clear old path table
	clear_path_table();

	// first, add own direct neighbors
	for(port_nr = 0; port_nr < NODE_PORT_CNT; port_nr++) {
		if(Node.port[port_nr].connected == XTRUE) {
			node_id = XNet.Link_State_Table[0][port_nr];
			if((node_id != XNET_NODE_ID_UNDEFINED) && (node_id != Node.my_id)) {
				// node is directly connected
				if(is_node_id_in_NodeIdList(node_id) == XFALSE) {
					// add this new node into NodeIdList
					add_node_id(node_id);
				}
				// set entry with hop = 1
				add_path_table_entry(node_id, port_nr, 1);
			}
		}
	}

	// second, add all indirect connected neighbors
	hop_cnt = 0;
	do {
		next_hop = XFALSE;
		hop_cnt++;
		for(n = 1; n < XNET_MAX_NODES; n++) {
			hops = XNet.Path_Table[n].hop_cnt;
			if(hops == hop_cnt) {
				// node found with matching hop counter
				port_nr = XNet.Path_Table[n].port_nr;
				for(m = 0; m < NODE_PORT_CNT; m++) {
					node_id = XNet.Link_State_Table[n][m];
					if((node_id != XNET_NODE_ID_UNDEFINED) && (node_id != Node.my_id)) {
						// node is connected to another node
						if(is_node_id_in_NodeIdList(node_id) == XFALSE) {
							// add this new node into NodeIdList
							add_node_id(node_id);
						}
						if(add_path_table_entry(node_id, port_nr, (hop_cnt+1)) == XTRUE) {
							// we had made a new entry, so we must check the table again
							next_hop = XTRUE;
						}
					}
				}
			}
		}
	}while(next_hop == XTRUE);

	// third, count and list all reachable nodes in the network
	XNet.connection.nr_nodes = 1;
	for(n = 1; n < XNET_MAX_NODES; n++) {
		if(XNet.Path_Table[n].hop_cnt != XNET_HOPS_UNDEFINED) {
			XNet.connection.nr_nodes++;
		}
	}

	Node.dbg_cnt.calc_PT++;
	UB_Xnet_L1_AddDbgLst(DBG_LIST_CALC_PT, XNet.connection.nr_nodes, 0, 0, 0, 0, 0, 0,0,0);
}

// new own LinkState status
void new_own_link_state(void)
{
	// update own link state list
	update_own_link_state();

	if(XNet.connection.nr_neighbors == 0) {
		clear_path_table();
		XNet.state.path_table_ok = XFALSE;
		XNet.connection.nr_nodes = 1;
	}
	else {
		// set flag to flood own link state
		Node.l2_flags.flags |= L2_FLAG_FLOOD_OWN_LINK_STATE;
	}

	Node.dbg_cnt.update_ls++;
}

// update own LinkStateList, LinkStateTable and Info
void update_own_link_state(void)
{
	U8_CHAR n, neighbor_id;

	XNet.connection.nr_neighbors = 0;

	for(n = 0; n < NODE_PORT_CNT; n++) {
		if(Node.port[n].connected == XFALSE) {
			// no neighbor on this port
			// set entry in LinkStateTable
			XNet.Link_State_Table[0][n] = XNET_NODE_ID_UNDEFINED;
		}
		else if(Node.port[n].neighbor_id != XNET_NODE_ID_UNDEFINED) {
			// neighbor on this port
			XNet.connection.nr_neighbors++;
			neighbor_id = Node.port[n].neighbor_id;
			// check if this a new neighbor on the port
			if((XNet.Link_State_Table[0][n] != neighbor_id) && (XNet.state.flooding_table == XFALSE)) {
				// check if the new neighbor already in our network
				if(is_node_in_my_network(neighbor_id) == XFALSE) {
					// store new neighbor data (if we have a higher node_id) as the neighbor
					if(Node.my_id > neighbor_id) {
						// to prevent neighbor to send also his table
						XNet.state.new_neighbor = XTRUE;
						XNet.state.neighbor_port = n;
					}
				}
			}
			// set entry in LinkStateTable
			XNet.Link_State_Table[0][n] = neighbor_id;
		}
	}
	UB_Xnet_L1_AddDbgLst(DBG_LIST_UPD_LST, XNet.connection.nr_neighbors, XNet.Link_State_Table[0][0], XNet.Link_State_Table[0][1], XNet.Link_State_Table[0][2], XNet.Link_State_Table[0][3], 0, 0,0,0);
}
